/***********************************************************************************

	Copyright (C) 2007-2011 Ahmet Öztürk (aoz_2@yahoo.com)

	This file is part of Lifeograph.

	Lifeograph is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Lifeograph is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

***********************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cstdio>	// for file operations
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include <gcrypt.h>
#include <cerrno>
#include <cassert>

#include "diary.hpp"
#include "helpers.hpp"
#include "lifeobase.hpp"


using namespace LIFEO;


// STATIC MEMBERS
Diary                   *Diary::d;
ElementShower< Diary >  *Diary::shower( NULL );

static const char	DEFAULT_CHAPTER_CTG_NAME[]	= N_( "Default" );

// PARSING HELPERS
Date
get_db_line_date( const Glib::ustring &line )
{
	Date date( 0 );

	for( unsigned int i = 2;
		 i < line.size() && i < 12 && int ( line[ i ] ) >= '0' && int ( line[ i ] ) <= '9';
		 i++ )
	{
		date.m_date = ( date.m_date * 10 ) + int ( line[ i ] ) - '0';
	}

	return( date );
}

Glib::ustring
get_db_line_name( const Glib::ustring &line )
{
	Glib::ustring::size_type begin( line.find( '\t' ) );
	if( begin == std::string::npos )
		begin = 2;
	else
		begin++;

	return( line.substr( begin ) );
}

// DIARY ===========================================================================================
Diary::Diary( void )
:   DiaryElement( this ), m_path( "" ), m_passphrase( "" ),
    m_ptr2chapter_ctg_cur( NULL ), m_topics( this, "topics" ),
    m_default_theme( ThemeSystem::get() ),
    m_startup_elem( HOME_CURRENT_ELEM ), m_last_elem( DEID_MIN ),
    m_option_sorting_criteria( SC_DATE ), m_read_version( 0 ),
#ifdef LIFEOGRAPH_DEBUG_BUILD
    m_flag_ignore_locks( true ),
#else
    m_flag_ignore_locks( false ),
#endif
    m_flag_read_only( false ),
    m_filter_text( "" ), m_filter_tag( NULL ), m_filter_date_begin( 0 ),
    m_filter_date_end( Date::DATE_MAX ), m_filtering_status( FS_CLEAR )
{
    // system theme added back to m_themes:
    m_themes.insert( PoolThemes::value_type( STRING::SYSTEM_THEME, ThemeSystem::get() ) );
}

Diary::~Diary( void )
{
	remove_lock();
}

void
Diary::init_new( const std::string &path )
{
    clear();
    set_path( path, true );
    add_today();
    // every diary must at least have one chapter category:
    m_ptr2chapter_ctg_cur = create_chapter_ctg( _( DEFAULT_CHAPTER_CTG_NAME ) );
}

void
Diary::clear( void )
{
    if( remove_lock() )
        m_path.clear();

	m_read_version = 0;
	DiaryBase::clear();			// clear IDs
	create_new_id( this );		// add DEID_MIN back to IDs pool
	create_new_id( &m_topics );	// FIXME: do topics really need a deid?
	m_entries.clear();
	m_tags.clear();
	m_tag_categories.clear();
	m_chapter_categories.clear();
	m_topics.clear();
	m_themes.clear();

	m_ptr2chapter_ctg_cur = NULL;
	m_default_theme = ThemeSystem::get();
	m_startup_elem = HOME_CURRENT_ELEM;
	m_last_elem = DEID_MIN;

	m_passphrase.clear();

	m_filter_text.clear();
	m_filter_tag = NULL;
    m_filter_date_begin = 0;
    m_filter_date_end = Date::DATE_MAX;
    m_filtering_status = FS_CLEAR;

	// NOTE: only reset body options here:
	m_language.clear();
	m_option_sorting_criteria = SC_DATE;
}

const Glib::RefPtr< Gdk::Pixbuf >&
Diary::get_icon( void ) const
{
	return Lifeobase::icons->diary_16;
}
const Glib::RefPtr< Gdk::Pixbuf >&
Diary::get_icon32( void ) const
{
    return Lifeobase::icons->diary_32;
}

LIFEO::Result
Diary::set_path( const std::string &path, bool new_file, bool read_only )
{
    // CHECK FILE SYSTEM PERMISSIONS
    if( access( path.c_str(), F_OK ) != 0 ) // check existence
    {
        if( errno == ENOENT )
        {
            if( !new_file )
            {
                PRINT_DEBUG( "File is not found" );
                return LIFEO::FILE_NOT_FOUND;
            }
        }
        else
            return LIFEO::FAILURE; // should not be the case
    }
    else if( access( path.c_str(), R_OK ) != 0 ) // check read access
    {
        PRINT_DEBUG( "File is not readable" );
        return LIFEO::FILE_NOT_READABLE;
    }
    else if( access( path.c_str(), W_OK ) != 0 ) // check write access
    {
        // errno == EACCES or errno == EROFS but no need to bother
        PRINT_DEBUG( "File is not writable, opening read-only..." );
        read_only = true;
    }

    // CHECK LOCK
    std::string path_lock( path + LOCK_SUFFIX );
    if( ! new_file && Glib::file_test( path_lock, Glib::FILE_TEST_EXISTS ) )
    {
        if( m_flag_ignore_locks )
            print_info( "Ignored file lock" );
        else
        if( ! read_only )
            return LIFEO::FILE_LOCKED;
    }

    // REMOVE PREVIOUS LOCK IF ANY
    if( ! new_file ) // init_new() clears diary first, so not necessary there
        remove_lock(); // remove existing lock

    // ACCEPT PATH
    m_path = path;
    m_name = Glib::filename_display_basename( path );
    m_flag_read_only = read_only;

	// "TOUCH" THE NEW LOCK
    if( ! read_only && ! new_file )
    {
        FILE *fp = fopen( path_lock.c_str(), "a+" );
        if( fp )
            fclose( fp );
    }

    return LIFEO::SUCCESS;
}

const std::string&
Diary::get_path( void ) const
{
	return m_path;
}

bool
Diary::is_path_set( void ) const
{
	return ( (bool) m_path.size() );
}

bool
Diary::set_passphrase( const std::string &passphrase )
{
	if( passphrase.size() >= PASSPHRASE_MIN_SIZE )
	{
		m_passphrase = passphrase;
		return true;
	}
	else
		return false;
}

void
Diary::clear_passphrase( void )
{
	m_passphrase.clear();
}

std::string
Diary::get_passphrase( void ) const
{
	return m_passphrase;
}

bool
Diary::compare_passphrase( const std::string &passphrase ) const
{
	return( m_passphrase == passphrase );
}

bool
Diary::is_passphrase_set( void ) const
{
	return (bool) m_passphrase.size();
}

LIFEO::Result
Diary::read_header()
{
	std::ifstream file( m_path.c_str() );

	if( ! file.is_open() )
	{
		print_error( "failed to open diary file " + m_path );
		return LIFEO::COULD_NOT_START;
	}
	std::string line;

	getline( file, line );

	if( line != LIFEO::DB_FILE_HEADER )
	{
		file.close();
		return LIFEO::CORRUPT_FILE;
	}

	while( getline( file, line ) )
	{
		switch( line[ 0 ] )
		{
			case 'V':
				m_read_version = convert_string( line.substr( 2 ) );
				if( m_read_version < LIFEO::DB_FILE_VERSION_INT_MIN ||
					m_read_version > LIFEO::DB_FILE_VERSION_INT )
				{
					file.close();
					return LIFEO::INCOMPATIBLE_FILE;
				}
				break;
			case 'E':
				if( line[ 2 ] == 'y' )
					// passphrase is set to a dummy value to indicate that diary
					// is an encrypted one until user enters the real passphrase
					m_passphrase = " ";
				else
					m_passphrase.clear();
				break;
			case 0: // end of header
				m_body_position = file.tellg();
				file.close();
				return LIFEO::SUCCESS;
			default:
				print_error( "unrecognized header line: " + line );
				break;
		}
	}

	file.close();
	m_path = "";
	return LIFEO::CORRUPT_FILE;
}

LIFEO::Result
Diary::read_body()
{
	if( m_passphrase.empty() )
		return read_plain();
	else
		return read_encrypted();
}

LIFEO::Result
Diary::read_plain()
{
	std::ifstream file( m_path.c_str() );

	if( ! file.is_open() )
	{
		print_error( "failed to open diary file " + m_path );
		return LIFEO::COULD_NOT_START;
	}

	file.seekg( 0, std::ios::end );

	int readsize = file.tellg() - m_body_position - 1;

	if( readsize <= 0 )
	{
		print_info( "diary has no body" );
		file.close();
		return LIFEO::CORRUPT_FILE;
	}

	file.seekg( m_body_position );

	char *readbuffer = new char[ readsize + 1 ];
	// TODO: check validity

	file.read( readbuffer, readsize );

	readbuffer[ readsize ] = 0;   // terminating zero

	file.close();

	std::stringstream stream( readbuffer );
	LIFEO::Result result = m_read_version == 56 ? parse_db_body_text_56( stream ) :
										parse_db_body_text( stream );
	delete[] readbuffer;

	return result;
}

LIFEO::Result
Diary::read_encrypted()
{
	std::ifstream file( m_path.c_str(), std::ios::in | std::ios::binary );

	if( ! file.is_open() )
	{
		print_error( "Failed to open diary file" + m_path );
		return LIFEO::COULD_NOT_START;
	}
	file.seekg( m_body_position );

	std::string line, content;
	unsigned char *buffer;
	try {
		// Allocate memory for salt
		unsigned char *salt = new unsigned char[ LIFEO::Cipher::cSALT_SIZE ];
		if( ! salt )
			throw LIFEO::Error( "Unable to allocate memory for salt" );
		// Read salt value
		file.read( (char*) salt, LIFEO::Cipher::cSALT_SIZE );

		unsigned char *iv = new unsigned char[ LIFEO::Cipher::cIV_SIZE ];
		if( ! iv )
			throw LIFEO::Error( "Unable to allocate memory for IV" );
		// Read IV
		file.read( (char*) iv, LIFEO::Cipher::cIV_SIZE );

		unsigned char *key;
		LIFEO::Cipher::expand_key( m_passphrase.c_str(), salt, &key );

		// Calulate bytes of data in file
		size_t size = LIFEO::get_file_size( file ) - file.tellg();
		if( size <= 3 )
		{
			file.close();
			return LIFEO::CORRUPT_FILE;
		}
		buffer = new unsigned char[ size + 1 ];
		// TODO(<0.4.0): ex handling here!!!

		file.read( (char*) buffer, size );
		LIFEO::Cipher::decrypt_buffer( buffer, size, key, iv );
		file.close();

		// passphrase check
		if( buffer[ 0 ] != m_passphrase[ 0 ] && buffer[ 1 ] != '\n' )
		{
			return LIFEO::WRONG_PASSWORD;
		}

		buffer[ size ] = 0;   // terminating zero
	}
	catch( ... )
	{
		return LIFEO::COULD_NOT_START;
	}

	std::stringstream stream;
	buffer += 2;    // ignore first two chars which are for passphrase checking
	stream << buffer;
	buffer -= 2;    // restore pointer to the start of the buffer before deletion
	delete[] buffer;

	return( m_read_version == 56 ? parse_db_body_text_56( stream ) :
						 parse_db_body_text( stream ) );
}

LIFEO::Result
Diary::parse_db_body_text( std::stringstream &stream )
{
	std::string			line( "" );
	Entry				*entry_new = NULL;
	CategoryChapters	*ptr2chapter_ctg = NULL;
	Chapter				*ptr2chapter = NULL;
	CategoryTags		*ptr2tag_ctg = NULL;
	Theme				*ptr2theme = NULL;
	bool				flag_first_paragraph( false );

	// TAG DEFINITIONS & CHAPTERS
	while( getline( stream, line ) )
	{
		if( line[ 0 ] == 0 )	// end of section
			break;
		else
		if( line.size() < 3 )
			continue;
		else
		{
			switch( line[ 0 ] )
			{
				case 'I':
					set_force_id( LIFEO::convert_string( line.substr( 2 ) ) );
				break;
				case 'T':	// tag category
					ptr2tag_ctg = create_tag_ctg( line.substr( 2 ) );
					ptr2tag_ctg->set_expanded( line[ 1 ] == 'e' );
					break;
				case 't':	// tag
					create_tag( line.substr( 2 ), ptr2tag_ctg );
					break;
				case 'C':	// chapter category
					ptr2chapter_ctg = create_chapter_ctg( line.substr( 2 ) );
					if( line[ 1 ] == 'c' || m_ptr2chapter_ctg_cur == NULL )	// for v73
						m_ptr2chapter_ctg_cur = ptr2chapter_ctg;
					break;
				case 'o':	// ordinal chapter (topic)
					ptr2chapter = m_topics.create_chapter(
							get_db_line_name( line ), get_db_line_date( line ) );
					ptr2chapter->set_expanded( line[ 1 ] == 'e' );
					break;
				case 'c':	// chapter
					if( ptr2chapter_ctg == NULL )
					{
						print_error( "No chapter category defined" );
						break;
					}
						if( m_read_version < 74 )
							ptr2chapter = new Chapter( this, line.substr( 2 ) );
						else
							ptr2chapter = ptr2chapter_ctg->create_chapter(
								get_db_line_name( line ), get_db_line_date( line ) );
						ptr2chapter->set_expanded( line[ 1 ] == 'e' );
					break;
				case 'd':	// chapter begin date (v73)
					if( ptr2chapter != NULL )	// chapter begin date
					{
						Date date( LIFEO::convert_string( line.substr( 2 ) ) );
						ptr2chapter_ctg->set_chapter_date( ptr2chapter, date );
					}
					break;
				case 'M':
					ptr2theme = create_theme( line.substr( 2 ) );
					if( line[ 1 ] == 'd' )
						m_default_theme = ptr2theme;
					break;
				case 'm':
					if( ptr2theme == NULL )
					{
						print_error( "No theme declared" );
						break;
					}
					switch( line[ 1 ] )
					{
						case 'f':	// font
							ptr2theme->font = Pango::FontDescription( line.substr( 2 ) );
							break;
						case 'b':	// base color
							ptr2theme->color_base.set( line.substr( 2 ) );
							break;
						case 't':	// text color
							ptr2theme->color_text.set( line.substr( 2 ) );
							break;
						case 'h':	// heading color
							ptr2theme->color_heading.set( line.substr( 2 ) );
							break;
						case 's':	// subheading color
							ptr2theme->color_subheading.set( line.substr( 2 ) );
							break;
						case 'l':	// highlight color
							ptr2theme->color_highlight.set( line.substr( 2 ) );
							break;
					}
					break;
				case 'O':	// options
					if( line.size() < 4 )
						break;
                    if ( line[ 2 ] == 's' ) // for versions prior to 110
                        m_language = HELPERS::get_env_lang();
					m_option_sorting_criteria = line[ 3 ];
					break;
                case 'l':   // language
                    m_language = line.substr( 2 );
                    break;
                case 'S':	// startup action
                    m_startup_elem = LIFEO::convert_string( line.substr( 2 ) );
                    break;
				case 'L':
					m_last_elem = LIFEO::convert_string( line.substr( 2 ) );
					break;
				default:
					print_error( "unrecognized line:\n" + line );
					return LIFEO::CORRUPT_FILE;
			}
		}
	}

	// ENTRIES
	while( getline( stream, line ) )
	{
		if( line.size() < 2 )
			continue;

		switch( line[ 0 ] )
		{
			case 'I':
				set_force_id( LIFEO::convert_string( line.substr( 2 ) ) );
				break;
			case 'E':	// new entry
            case 'e':   // trashed
				entry_new = new Entry( this, LIFEO::convert_string( line.substr( 2 ) ),
									   line[ 1 ] == 'f' );
				if( m_read_version < 74 )
					entry_new->m_date.m_date++;	// entry orders should start from 1
				m_entries[ entry_new->m_date.m_date ] = entry_new;
                if( line[ 0 ] == 'e' )
                {
                    entry_new->set_trashed( true );
                    // trashed entries are always hidden at the login
                    entry_new->set_filtered_out( true );
                }
				flag_first_paragraph = true;
				break;
			case 'D':	// creation & change dates (optional)
				if( entry_new == NULL )
				{
					print_error( "No entry declared" );
					break;
				}
				if( line[ 1 ] == 'r' )
					entry_new->m_date_created = LIFEO::convert_string( line.substr( 2 ) );
				else	// it should be 'h'
					entry_new->m_date_changed = LIFEO::convert_string( line.substr( 2 ) );
				break;
			case 'M':	// theme
				if( entry_new == NULL )
					print_error( "No entry declared" );
				else
				{
					PoolThemes::iterator iter_theme( m_themes.find( line.substr( 2 ) ) );
					if( iter_theme != m_themes.end() )
						entry_new->set_theme( iter_theme->second );
					// system theme added back to m_themes:
					//else if( line.substr( 2 ) == STRING::SYSTEM_THEME )
					//    entry_new->set_theme( ThemeSystem::get() );
					else
						print_error( "Reference to undefined theme: " + line.substr( 2 ) );
				}
				break;
			case 'T':	// tag
				if( entry_new == NULL )
					print_error( "No entry declared" );
				else
				{
					PoolTags::iterator iter_tag( m_tags.find( line.substr( 2 ) ) );
					if( iter_tag != m_tags.end() )
						entry_new->add_tag( iter_tag->second );
					else
						print_error( "Reference to undefined tag: " + line.substr( 2 ) );
				}
				break;
            case 'l':   // language
                if( entry_new == NULL )
                    print_error( "No entry declared" );
                else
                    entry_new->set_lang( line.substr( 2 ) );
                break;
			case 'P':    // paragraph
				if( entry_new == NULL )
				{
					print_error( "No entry declared" );
					break;
				}
				if( flag_first_paragraph )
				{
					if( line.size() > 2 )
						entry_new->m_text = line.substr( 2 );
					entry_new->m_name = entry_new->m_text;
					flag_first_paragraph = false;
				}
				else
				{
					entry_new->m_text += "\n";
					entry_new->m_text += line.substr( 2 );
				}
				break;
			default:
				print_error( "unrecognized line:\n" + line );
				return LIFEO::CORRUPT_FILE;
				break;
		}
	}

    // every diary must at least have one chapter category:
    if( m_chapter_categories.empty() )
        m_ptr2chapter_ctg_cur = create_chapter_ctg( _( DEFAULT_CHAPTER_CTG_NAME ) );

	if( m_startup_elem > HOME_FIXED_ELEM )
		if( get_element( m_startup_elem ) == NULL )
		{
			print_error( "startup element cannot be found in db" );
			m_startup_elem = DEID_MIN;
		}

	if( m_entries.size() < 1 )
	{
		print_info( "a dummy entry added to the diary" );
		add_today();
	}

	return LIFEO::SUCCESS;
}

LIFEO::Result
Diary::parse_db_body_text_56( std::stringstream &stream )
{
	std::string			line( "" );
	Glib::ustring		content( "" );
	Date::date_t		date( 0 );
	bool				flag_favorite( false );
	Entry				*entry_new = NULL;
	CategoryChapters	*ptr2chapter_ctg = NULL;
	Chapter				*ptr2chapter = NULL;
	std::vector< Tag* >	tmp_vector_tags;
	std::string			str_tags;

	// TAG DEFINITIONS & CHAPTERS
	while( getline( stream, line ) )
	{
		if( line[ 0 ] == 0 )
			break;
		else
		if( line.size() < 3 )
			continue;
		else
		{
			switch( line[ 0 ] )
			{
				case 'T':	// tag
					try
					{
						Tag *tmp_tag = create_tag( line.substr( 2 ) );
						tmp_vector_tags.push_back( tmp_tag );
						// TODO: this is a temporary solution. will be fixed in new diary version
					}
					catch( std::exception &ex )
					{
						throw ex;
					}
					break;
				case 'L':	// current entryliststyle
					// not used anymore...
					break;
				case 'S':	// chapterset
					try
					{
						ptr2chapter_ctg = create_chapter_ctg( line.substr( 2 ) );
						if( m_ptr2chapter_ctg_cur == NULL )
							m_ptr2chapter_ctg_cur = ptr2chapter_ctg;
					}
					catch( std::exception &ex )
					{
						throw ex;
					}
					break;
				case 'C':	// chapter
					if( m_chapter_categories.size() > 0 )
					{
						try
						{
							ptr2chapter = new Chapter( this, line.substr( 2 ) );
						}
						catch( std::exception &ex )
						{
							throw ex;
						}
					}
					break;
				case 'd':	// begin date
					if( ptr2chapter != NULL )	// chapter begin date
					{
						ptr2chapter_ctg->set_chapter_date( ptr2chapter, Date( line.substr( 2 ) ) );
					}
					break;
				case 'O':	// options
                    if ( line[ 2 ] == 's' )
                        m_language = HELPERS::get_env_lang();
					break;
				case 'M':
					if( m_default_theme == ThemeSystem::get() )
					{
						m_default_theme = create_theme( _( "theme" ) );
					}

					switch( line[1] )
					{
						case 'f':	// font
							m_default_theme->font = Pango::FontDescription( line.substr( 2 ) );
							break;
						case 'b':	// base color
							m_default_theme->color_base.set( line.substr( 2 ) );
							break;
						case 't':	// text color
							m_default_theme->color_text.set( line.substr( 2 ) );
							break;
						case 'h':	// heading color
							m_default_theme->color_heading.set( line.substr( 2 ) );
							break;
						case 's':	// subheading color
							m_default_theme->color_subheading.set( line.substr( 2 ) );
							break;
						case 'l':	// highlight color
							m_default_theme->color_highlight.set( line.substr( 2 ) );
							break;
					}
					break;
				default:
					print_error( "unrecognized line:\n" + line );
					return LIFEO::CORRUPT_FILE;
			}
		}
	}

	// ENTRIES
	while( getline( stream, line ) )
	{
		switch( line[ 0 ] )
		{
			case 'A':	// date
			{	// FIXME: temporary (for compatibility)
				Date::date_t date_new = LIFEO::convert_string( line.substr( 2 ) );
				if( date_new == ( date >> 10 ) )
					date++;
				else
					date = ( date_new << 10 ) + 1;
				break;
			}
			case 'C':    // content
				// when writing a diary file by hand, one can easily forget to put
				// a blank space after C, so we have to check
				if( line.size() < 2 )
					continue;

				content += line.substr( 2 );
				content += "\n";
				break;
			case 'T':    // tags
				str_tags = line;
			break;
			case '}':    // end of Entry
			{
				// erase redundant new line:
				content.erase( content.size() - 1 );

				entry_new = new Entry( this, date, content, flag_favorite );

				// tags
				std::vector< Tag* >::const_iterator iter_tag = tmp_vector_tags.begin();
				for( std::string::size_type i = 2; i < str_tags.size(); i++ )
				{
					if( iter_tag == tmp_vector_tags.end() )
						//TODO: issue a warning
						break;
					else
					if( str_tags[ i ] == '#' )
						entry_new->add_tag( *iter_tag );
					else
					if( str_tags[ i ] == ' ' )
						continue;

					++iter_tag;
				}
				// TODO: per entry spellcheck options
				//entry_new->set_spellcheck( opt_spell );
				m_entries[ date ] = entry_new;

				content.clear();
				str_tags.clear();
				break;
			}
			case '{':
				flag_favorite = false;
				break;
			case '[':
				flag_favorite = true;
				break;
			case 0:
			case '\n':
				break;
			default:
				print_error( "unrecognized line:\n" + line );
				return LIFEO::CORRUPT_FILE;
				break;
		}
	}

    // every diary must at least have one chapter category:
    if( m_chapter_categories.empty() )
        m_ptr2chapter_ctg_cur = create_chapter_ctg( _( DEFAULT_CHAPTER_CTG_NAME ) );

    if( m_entries.size() < 1 )
    {
        print_info( "a dummy entry added to the diary" );
        add_today();
    }

    return LIFEO::SUCCESS;
}

LIFEO::Result
Diary::write( void )
{
    assert( m_flag_read_only == false );

    m_flag_only_save_filtered = false;

	std::string path_old( m_path + ".~previousversion~" );

	if( Glib::file_test( m_path, Glib::FILE_TEST_EXISTS ) )
		rename( m_path.c_str(), path_old.c_str() );

	Result result( write( m_path ) );

// DAILY BACKUP SAVES
#ifdef LIFEOGRAPH_DEBUG_BUILD
    if( ! m_entries.empty() )
    {
        Glib::Date gdate;
        gdate.set_time_current();
        Date date( gdate.get_year(), gdate.get_month(), gdate.get_day() );

        std::string path_daily( m_path + "." + date.format_string() );

        if( write( path_daily ) == SUCCESS )
            print_info( "daily backup has been written successfully" );
    }
#endif

	return result;
}

LIFEO::Result
Diary::write( const std::string &path )
{
    if( m_passphrase.empty() )
        return write_plain( path );
    else
        return write_encrypted( path );
}

LIFEO::Result
Diary::write_copy( const std::string &path, const std::string &passphrase, bool flag_filtered )
{
    m_flag_only_save_filtered = flag_filtered;

	Result result;

	if( passphrase.empty() )
		result = write_plain( path );
	else
	{	
		std::string passphrase_actual( m_passphrase );
		m_passphrase = passphrase;
		result = write_encrypted( path );
		m_passphrase = passphrase_actual;
	}

	return result;
}

LIFEO::Result
Diary::write_txt( const std::string &path, bool flag_filtered )
{
	std::ofstream file( path.c_str(), std::ios::out | std::ios::trunc );
	if( ! file.is_open() )
	{
		print_error( "i/o error: " + path );
		return LIFEO::COULD_NOT_START;
	}

    const CategoryChapters  *chapters( Diary::d->get_current_chapter_ctg() );
    CategoryChapters::const_reverse_iterator
                            iter_chapter( chapters->rbegin() );
    Chapter                 *chapter( NULL );
    bool                    flag_chapter_round( true );    // chapter or topic round
    const std::string       separator( "---------------------------------------------\n");

	// ENTRIES
    for( EntryIterReverse itr_entry = m_entries.rbegin();
         itr_entry != m_entries.rend();
         ++itr_entry )
	{
        Entry *entry( itr_entry->second );

        for( ; ; ++iter_chapter)
        {
            if( iter_chapter == chapters->rend() )
            {
                if( flag_chapter_round )
                {
                    chapters = Diary::d->get_topics();
                    iter_chapter = chapters->rbegin();
                    if( iter_chapter == chapters->rend() )
                        break;
                    flag_chapter_round = false;
                    continue;
                }
                else
                    break;
            }

            chapter = iter_chapter->second;

            if( chapter->is_initialized() )
            {
                if( entry->get_date() >= chapter->get_date() )
                    file << "\n\n" << separator << chapter->get_date().format_string() <<
                    " - " << chapter->get_name() << "\n";
                else
                    break;
            }
        }

        // purge empty entries:
        if( ( entry->m_text.size() < 1 && entry->get_tags().size() < 1 ) ||
            ( entry->get_filtered_out() && flag_filtered ) )
            continue;

        file << separator;

        // DATE AND FAVOREDNESS
        file << entry->get_date().format_string();
        if( entry->is_favored() )
            file << "  [" << _( "FAVORITE" ) << "]";

        // CONTENT
        file << "\n" << entry->get_text();

        // TAGS
        const Tagset &tagset = entry->get_tags();
        for( Tagset::const_iterator itr_tags = tagset.begin();
             itr_tags != tagset.end();
             ++itr_tags )
        {
            if( itr_tags == tagset.begin() )
                file << "\n\n" << _( "TAGS" ) << ": ";
            else
                file << ", ";

            file << ( *itr_tags )->get_name();
        }

        file << "\n\n\n";
    }
    file.close();
    return SUCCESS;
}

LIFEO::Result
Diary::write_plain( const std::string &path, bool flag_header_only )
{
	std::ofstream file( path.c_str(), std::ios::out | std::ios::trunc );
	if( ! file.is_open() )
	{
		print_error( "i/o error: " + path );
		return LIFEO::COULD_NOT_START;
	}

	std::stringstream output;
	create_db_header_text( output, flag_header_only );
	// header only mode is for encrypted diaries
	if( ! flag_header_only )
	{
		create_db_body_text( output );
	}

	file << output.str();
	file.close();

	return LIFEO::SUCCESS;
}

LIFEO::Result
Diary::write_encrypted( const std::string &path )
{
	// writing header:
	write_plain( path, true );
	std::ofstream file( path.c_str(), std::ios::out | std::ios::app | std::ios::binary );
	if( ! file.is_open() )
	{
		print_error( "i/o error: " + path );
		return LIFEO::COULD_NOT_START;
	}
	std::stringstream output;
	// first char of passphrase for validity checking
	output << m_passphrase[ 0 ] << '\n';
	create_db_body_text( output );

	// encryption
	try {
		size_t size =  output.str().size() + 1;

		unsigned char *key, *salt;
		LIFEO::Cipher::create_new_key( m_passphrase.c_str(), &salt, &key );

		unsigned char *iv;
		LIFEO::Cipher::create_iv( &iv );

		unsigned char *buffer = new unsigned char[ size ];
		memcpy( buffer, output.str().c_str(), size );

		LIFEO::Cipher::encrypt_buffer( buffer, size, key, iv );

		file.write( (char*) salt, LIFEO::Cipher::cSALT_SIZE );
		file.write( (char*) iv, LIFEO::Cipher::cIV_SIZE );
		file.write( (char*) buffer, size );

	}
	catch( ... )
	{
		return LIFEO::FAILURE;
	}

	file.close();

	return LIFEO::SUCCESS;
}

bool
Diary::create_db_header_text( std::stringstream &output, bool encrypted )
{
	output << LIFEO::DB_FILE_HEADER;
	output << "\nV " << LIFEO::DB_FILE_VERSION_INT;
	output << ( encrypted ? "\nE yes" : "\nE no" );
	output << "\n\n"; // end of header

	return true;
}

bool
Diary::create_db_body_text( std::stringstream &output )
{
	// OPTIONS
    // dashed char used to be used for spell-checking before v110
    output << "O -" << m_option_sorting_criteria << '\n';
    if( !m_language.empty() )
        output << "l " << m_language << '\n';

	// STARTUP ACTION (HOME ITEM)
	output << "S " << m_startup_elem << '\n';
	output << "L " << m_last_elem << '\n';

	// ROOT TAGS
	for( PoolTags::const_iterator iter = m_tags.begin();
		 iter != m_tags.end();
		 ++iter )
	{
		if( iter->second->get_category() == NULL )
            output << "ID" << iter->second->get_id()
                   << "\nt " << iter->second->get_name_std() << '\n';
	}
	// CATEGORIZED TAGS
	for( PoolCategoriesTags::const_iterator iter = m_tag_categories.begin();
		 iter != m_tag_categories.end();
		 ++iter )
	{
        // tag category:
		CategoryTags *ctg( iter->second );
		output << "ID" << ctg->get_id()
               << "\nT" << ( ctg->get_expanded() ? 'e' : ' ' )
               << ctg->get_name_std() << '\n';
		// tags in it:
		for( CategoryTags::const_iterator iter_tag = ctg->begin();
			 iter_tag != ctg->end();
			 ++iter_tag )
		{
            output << "ID" << ( *iter_tag )->get_id()
                   << "\nt "<< ( *iter_tag )->get_name_std() << '\n';
		}
	}
	// TOPICS
	for( CategoryChapters::iterator iter_chapter = m_topics.begin();
		 iter_chapter != m_topics.end();
		 ++iter_chapter )
	{
        output << "ID" << iter_chapter->second->get_id()
               << "\no" <<	( iter_chapter->second->get_expanded() ? 'e' : ' ' )
               << iter_chapter->first   // date (order)
               << '\t' << iter_chapter->second->get_name_std() << '\n';
	}
	// CHAPTERS
	for( PoolCategoriesChapters::iterator iter = m_chapter_categories.begin();
		 iter != m_chapter_categories.end();
		 ++iter )
	{
        // chapter category:
		CategoryChapters *ctg( iter->second );
        output << "ID" << ctg->get_id()
               << "\nC" << ( ctg == m_ptr2chapter_ctg_cur ? 'c' : ' ' )
               << ctg->get_name_std() << '\n';
        // chapters in it:
		for( CategoryChapters::iterator iter_chapter = ctg->begin();
			 iter_chapter != ctg->end();
			 ++iter_chapter )
		{
            output << "ID" << iter_chapter->second->get_id()
                   << "\nc" << ( iter_chapter->second->get_expanded() ? 'e' : ' ' )
                   << iter_chapter->first   // date
                   << '\t' << iter_chapter->second->get_name_std() << '\n';
		}
	}
	// THEMES
	for( PoolThemes::const_iterator iter = m_themes.begin();
		 iter != m_themes.end();
		 ++iter )
	{
		Theme *theme( iter->second );

		if( theme->is_system() )
			continue;

        output << "ID" << theme->get_id()
               << "\nM" << ( theme == m_default_theme ? 'd' : ' ' )
               << theme->get_name_std() << '\n';
        output << "mf" << theme->font.to_string() << '\n';
        output << "mb" << convert_gdkrgba_to_string( theme->color_base ) << '\n';
        output << "mt" << convert_gdkrgba_to_string( theme->color_text ) << '\n';
        output << "mh" << convert_gdkrgba_to_string( theme->color_heading ) << '\n';
        output << "ms" << convert_gdkrgba_to_string( theme->color_subheading ) << '\n';
        output << "ml" << convert_gdkrgba_to_string( theme->color_highlight ) << '\n';
	}

	output << '\n'; // end of section

	// ENTRIES
	for( Entryiter iter_entry = m_entries.begin();
		 iter_entry != m_entries.end();
		 ++iter_entry )
	{
		Entry *entry = ( *iter_entry ).second;

		// purge empty entries:
		if( entry->m_text.size() < 1 && entry->m_tags.empty() ) continue;
		// optionally only save filtered entries:
		else
		if( entry->get_filtered_out() && m_flag_only_save_filtered ) continue;

		// ENTRY DATE
        output << "ID" << entry->get_id() << "\n"
               << ( entry->is_trashed() ? 'e' : 'E' )
               << ( entry->m_option_favored ? 'f' : ' ' )
               << entry->m_date.m_date << '\n';
        output << "Dr" << entry->m_date_created << '\n';
        output << "Dh" << entry->m_date_changed << '\n';

		// THEME
		if( entry->get_theme_is_set() )
			output << "M " << entry->get_theme()->get_name_std() << '\n';

		// TAGS
		for( Tagset::const_iterator iter_tag = entry->m_tags.begin();
			 iter_tag != entry->m_tags.end();
			 ++iter_tag )
			output << "T " << ( *iter_tag )->get_name_std() << '\n';

        // LANGUAGE
        if( entry->get_lang() != LANG_INHERIT_DIARY )
            output << "l " << entry->get_lang() << '\n';

		// CONTENT
		if( entry->m_text.empty() )
            output << "\n";
		else
		{
            // NOTE: for some reason, implicit conversion from Glib:ustring...
            // ...fails while substr()ing when LANG=C
            // we might reconsider storing text of entries as std::string.
            // for now we convert entry text to std::string here:
            std::string             content( entry->m_text );
            std::string::size_type  pt_start( 0 ), pt_end( 0 );

            while( true )
            {
                pt_end = content.find( '\n', pt_start );
                if( pt_end == std::string::npos )
                {
                    pt_end = content.size();
                    output << "P " << content.substr( pt_start, content.size() - pt_start )
                           << "\n\n";
                    break; // end of while( true )
                }
                else
                {
                    pt_end++;
                    output << "P " << content.substr( pt_start, pt_end - pt_start );
                    pt_start = pt_end;
                }
            }
        }
    }

	return true;
}

DiaryElement*
Diary::get_element( DEID id )
{
	PoolDEIDs::iterator iter( m_ids.find( id ) );
	return( iter == m_ids.end() ? NULL : iter->second );
}

DiaryElement*
Diary::get_startup_elem( void )
{
	switch( m_startup_elem )
	{
		case HOME_CURRENT_ELEM:
			return get_most_current_elem();
		break;
		case HOME_LAST_ELEM:
			return get_element( m_last_elem );
		break;
		case DEID_UNSET:
			return NULL;
		default:
			return get_element( m_startup_elem );
	}
}

DiaryElement*
Diary::get_most_current_elem( void )
{
	Glib::Date gdate;
	gdate.set_time_current();
	Date date( gdate.get_year(), gdate.get_month(), gdate.get_day() );
	Date::date_t diff1( Date::ORDINAL_FLAG );
	long diff2;
	DiaryElement *elem( this );
	bool descending( false );
	for( Entryiter iter = m_entries.begin(); iter != m_entries.end(); ++iter )
	{
		if( ! iter->second->get_filtered_out() && ! iter->second->get_date().is_ordinal() )
		{
			diff2 = iter->second->get_date().m_date - date.m_date;
			if( diff2 < 0 ) diff2 *= -1;
			if( static_cast< unsigned long >( diff2 ) < diff1 )
			{
				diff1 = diff2;
				elem = iter->second;
				descending = true;
			}
			else
			if( descending )
				break;
		}
	}
	return elem;
}

DiaryElement*
Diary::get_prev_session_elem( void )
{
	return get_element( m_last_elem );
}

// LOCKS ===========================================================================================
inline bool
Diary::remove_lock( void )
{
    if( m_path.empty() )
        return false;

    std::string path_lock( m_path + LOCK_SUFFIX );
    remove( path_lock.c_str() );
    return true;
}

// ENTRIES =========================================================================================
Entry*
Diary::add_today( void )
{
    return create_entry( Date::get_today() );
}

Entry*
Diary::get_entry( const Date::date_t date )
{
    Entryiter iter( m_entries.find( date ) );
    if( iter != m_entries.end() )
        if( iter->second->get_filtered_out() == false )
            return iter->second;

	return NULL;
}

Entry*
Diary::get_entry_today( void )
{
    // FIXME: handle filtered out case
    return get_entry( Date::get_today( 1 ) );  // 1 is the order
}

Entryvector*
Diary::get_entries( Date::date_t date )	// takes pure date
{
	Entryvector *entries = new Entryvector;

	Entryiter iter = m_entries.find( date + 1 );
	if( iter == m_entries.end() )
		return entries;	// return empty vector

	for( ; ; --iter )
	{
		if( iter->second->get_date().get_pure() == date )
		{
			if( iter->second->get_filtered_out() == false )
				entries->push_back( iter->second );
		}
		else
			break;
		if( iter == m_entries.begin() )
			break;
	}
	return entries;
}

bool
Diary::get_day_has_multiple_entries( const Date &date_impure )
{
	Date::date_t date = date_impure.get_pure();
	Entryiter iter = m_entries.find( date + 2 );
	if( iter == m_entries.end() )
		return false;

	for( ; iter != m_entries.begin(); --iter )
	{
		if( iter->second->get_date().get_pure() == date )
		{
			if( iter->second->get_filtered_out() == false )
				return true;
		}
		else
			break;
	}

	return false;
}

Entry*
Diary::get_entry_next_in_day( const Date &date )
{
    Entryiter entry_1st( m_entries.find( date.get_pure() + 1 ) );
    if( entry_1st == m_entries.end() )
        return NULL;
    Entryiter entry_next( m_entries.find( date.m_date + 1 ) );
    if( entry_next != m_entries.end() )
        return entry_next->second->get_filtered_out() ? NULL : entry_next->second;
    else
    if( date.get_order() > 1 )
        return entry_1st->second->get_filtered_out() ? NULL : entry_1st->second;
    else
        return NULL;
}

Entry*
Diary::get_entry_first( void )
{
	// return first unfiltered entry
	for( Entryiter iter = m_entries.begin(); iter != m_entries.end(); ++iter )
	{
		Entry *entry = iter->second;
		if( entry->get_filtered_out() == false )
			return( entry );
	}
	return NULL;
}

bool
Diary::is_first( Entry const*const entry ) const
{
	return( entry == m_entries.begin()->second );
}

bool
Diary::is_last( Entry const*const entry ) const
{
	return( entry == m_entries.rbegin()->second );
}

Entry*
Diary::create_entry( Date::date_t date, const Glib::ustring &content, bool flag_favorite )
{
	// make it the last entry of its day:
	Date::reset_order_1( date );
	while( m_entries.find( date ) != m_entries.end() )
		++date;

	Entry *entry = new Entry( this, date, content, flag_favorite );

	m_entries[ date ] = entry;

	return( entry );
}

bool
Diary::dismiss_entry( Entry *entry )
{
	Date::date_t date = entry->get_date().m_date;

	// fix startup element
	if( m_startup_elem == entry->get_id() )
		m_startup_elem = DEID_MIN;

	// remove from tags:
	for( Tagset::iterator iter = entry->get_tags().begin();
		 iter != entry->get_tags().end();
		 ++iter )
		( *iter )->get_items()->erase( entry );

	// erase entry from map:
	m_entries.erase( date );

	// fix entry order:
	int i = 1;
	for( Entryiter iter = m_entries.find( date + i );
		 iter != m_entries.end();
		 iter = m_entries.find( date + i ) )
	{
		Entry *entry2fix = iter->second;
		m_entries.erase( iter );
		entry2fix->m_date.m_date--;
		m_entries[ entry2fix->get_date().m_date ] = entry2fix;
		++i;
	}

	delete entry;
	return true;
}

/*int
Diary::list_entries( void )
{
	int number_of_entries( 0 );
	for( Entryiter t1_itr = m_entries.begin(); t1_itr != m_entries.end(); t1_itr++ )
	{
		std::cout << t1_itr->second->get_date().format_string() << std::endl;
		number_of_entries++;
	}
	std::cout << number_of_entries << " entries total." << std::endl;
	return( number_of_entries );
}*/

void
Diary::set_entry_date( Entry *entry, const Date &date )
{
	Entryiter iter;
	Date::date_t d( entry->m_date.m_date );
	m_entries.erase( d );

	for( iter = m_entries.find( ++d );
		 iter != m_entries.end();
		 iter = m_entries.find( ++d ) )
	{
		Entry *entry_shift( iter->second );
		m_entries.erase( d );
		entry_shift->set_date( Date( d - 1 ) );
		m_entries[ d - 1 ] = entry_shift;
	}

	// find the last entry in the date/order
	d = date.m_date;
	bool flag_replace( false );
	while( m_entries.find( d ) != m_entries.end() )
	{
		flag_replace = true;
		d++;
	}

	if( flag_replace )
	{
		for( iter = m_entries.find( --d );
			 d >= date.m_date;
			 iter = m_entries.find( --d ) )
		{
			Entry *entry_shift( iter->second );
			m_entries.erase( d );
			entry_shift->set_date( Date( d + 1 ) );
			m_entries[ d + 1 ] = entry_shift;
		}
	}

	entry->set_date( date );
	m_entries[ date.m_date ] = entry;
}

// FILTERING =======================================================================================
void
Diary::set_filter_text( const Glib::ustring &filter )
{
	m_filter_text = filter;

    if( filter.empty() )
        m_filtering_status &= ( ~ FS_FILTER_TEXT );
    else
        m_filtering_status |= FS_FILTER_TEXT;

	m_filtering_status |= FS_NEW;
}

void
Diary::set_filter_tag( const Tag *tag )
{
	m_filter_tag = tag;

    if( tag == NULL )
        m_filtering_status &= ( ~FS_FILTER_TAG );
    else
        m_filtering_status |= FS_FILTER_TAG;

	m_filtering_status |= FS_NEW;
}

void
Diary::toggle_filter_favorites( void )
{
	if( m_filtering_status & FS_FILTER_FAVORITES )
		m_filtering_status -= FS_FILTER_FAVORITES;
	else
		m_filtering_status |= FS_FILTER_FAVORITES;

	m_filtering_status |= FS_NEW;
}

void
Diary::toggle_filter_trash( void )
{
    if( m_filtering_status & FS_FILTER_TRASH )
        m_filtering_status -= FS_FILTER_TRASH;
    else
        m_filtering_status |= FS_FILTER_TRASH;

    m_filtering_status |= FS_NEW;
}

void
Diary::clear_filter_date( void )
{
    m_filter_date_begin = 0;
    m_filter_date_end = Date::DATE_MAX;
    if( m_filtering_status & FS_FILTER_DATE )
        m_filtering_status -= FS_FILTER_DATE;

    m_filtering_status |= FS_NEW;
}

void
Diary::set_filter_date_begin( Date::date_t d )
{
    m_filtering_status |= FS_FILTER_DATE;
    m_filtering_status |= FS_NEW;
    m_filter_date_begin = d;
}

void
Diary::set_filter_date_end( Date::date_t d )
{
    m_filtering_status |= FS_FILTER_DATE;
    m_filtering_status |= FS_NEW;
    m_filter_date_end = d;
}

// called when some entries are filtered out individually irrespective of filtering criteria
void
Diary::activate_individual_filter( void )
{
    m_filtering_status |= FS_FILTER_DATE;
}

int
Diary::replace_text( const Glib::ustring &newtext )
{
	Glib::ustring::size_type iter_str;
	const int chardiff = newtext.size() - m_filter_text.size();

	Entry *entry;

	for(	Entryiter iter_entry = m_entries.begin();
			iter_entry != m_entries.end();
			iter_entry++ )
	{
		entry = iter_entry->second;
		if( entry->get_filtered_out() )
			continue;

		Glib::ustring entrytext = entry->get_text().lowercase();
		iter_str = 0;
		int count = 0;
		while(	( iter_str = entrytext.find( m_filter_text, iter_str ) ) !=
				std::string::npos )
		{
			entry->get_text().erase(
				iter_str + ( chardiff * count ), m_filter_text.size() );
			entry->get_text().insert(
				iter_str + ( chardiff * count ), newtext );
			count++;
			iter_str += m_filter_text.size();
		}
	}
	
	return 0;	// reserved
}

// TAGS ============================================================================================
Tag*
Diary::create_tag( const Glib::ustring &name, CategoryTags *category )
{
	PoolTags::iterator iter = m_tags.find( name );
	if( iter != m_tags.end() )
	{
		PRINT_DEBUG( "Tag already exists: " + name );
		return( iter->second );
	}
	Tag *tag( new Tag( this, name, category ) );
	m_tags.insert( PoolTags::value_type( name, tag ) );
	return tag;
}

void
Diary::dismiss_tag( Tag *tag )
{
	// remove from entries:
	for( SetDiaryElements::const_iterator iter = tag->get_items()->begin();
		 iter != tag->get_items()->end();
		 ++ iter )
	{
		Entry *entry = dynamic_cast< Entry* >( *iter );
		entry->remove_tag( tag );
	}

	// remove from category if any:
	if( tag->get_category() != NULL )
		tag->get_category()->erase( tag );

	// clear filter if necessary:
	if( tag == m_filter_tag )
		m_filter_tag = NULL;

	m_tags.erase( tag->get_name() );
	delete tag;
}

CategoryTags*
Diary::create_tag_ctg( void )
{
	Glib::ustring name = create_unique_name_for_map( m_tag_categories, _( "New category" ) );
	CategoryTags *new_category = new CategoryTags( this, name );
	m_tag_categories.insert( PoolCategoriesTags::value_type( name, new_category ) );

	return new_category;
}

CategoryTags*
Diary::create_tag_ctg( const Glib::ustring &name )	// used while reading diary file
{
	try
	{
		CategoryTags *new_category = new CategoryTags( this, name );
		m_tag_categories.insert( PoolCategoriesTags::value_type( name, new_category ) );
		return new_category;
	}
	catch( std::exception &ex )
	{
		PRINT_DEBUG( ex.what() );
		throw ex;
	}
}

void
Diary::dismiss_tag_ctg( CategoryTags *category )
{
	// fix entries
	for( CategoryTags::const_iterator iter = category->begin();
		 iter != category->end();
		 ++iter )
	{
		( * iter )->set_category( NULL );
	}
	// remove from the list and delete
	m_tag_categories.erase( category->get_name() );
	delete category;
}

// CHAPTERS ========================================================================================
CategoryChapters*
Diary::create_chapter_ctg( void )
{
	Glib::ustring name = create_unique_name_for_map( m_chapter_categories, _( "New category" ) );
	CategoryChapters *category = new CategoryChapters( this, name );
	m_chapter_categories.insert(
			PoolCategoriesChapters::value_type( name, category ) );

	return category;
}

CategoryChapters*
Diary::create_chapter_ctg( const Glib::ustring &name )
{
	// TODO: check string's availability
	try
	{
		CategoryChapters *category = new CategoryChapters( this, name );
		m_chapter_categories.insert(
				PoolCategoriesChapters::value_type( name, category ) );
		return category;
	}
	catch( std::exception &ex )
	{
		throw ex;
	}
}

void
Diary::dismiss_chapter_ctg( CategoryChapters *category )
{
	if( category == m_ptr2chapter_ctg_cur )
		m_ptr2chapter_ctg_cur = m_chapter_categories.begin()->second;

	m_chapter_categories.erase( category->get_name() );
    delete category;
}

void
Diary::dismiss_chapter( Chapter *chapter )
{
	// TODO: add an option to also delete entries within the chapter
	if( chapter->is_ordinal() )
	{
		// ORDER SHIFTING TO PRESERVE CONTINUITY
		if( m_topics.size() == 1 )	// must be handled separately
		{
			m_topics.erase( chapter->get_date().m_date );
			delete chapter;
			return;
		}

		// CALCULATE HELPER VALUES
		Date d_chapter( chapter->get_date() );
		Date::date_t d_chapter_next( d_chapter.m_date + Date::ORDINAL_STEP );
		bool flag_last_chapter( m_topics.find( d_chapter_next ) == m_topics.end() );

		int last_order_of_prev_chapter( 0 );
		if( flag_last_chapter )
		{
			Chapter *chapter_prev( m_topics.get_chapter(
					d_chapter.m_date - Date::ORDINAL_STEP ) );
			Date d_first_free( chapter_prev->get_free_order() );
			last_order_of_prev_chapter = d_first_free.get_order() - 1;
		}
		else
		{
			Date d_first_free( chapter->get_free_order() );
			last_order_of_prev_chapter = d_first_free.get_order() - 1;
		}

		// ACTUALLY DISMISS THE TOPIC
		m_topics.erase( chapter->get_date().m_date );
		delete chapter;

		// SHIFT TOPICS
		for( Date::date_t d = d_chapter.m_date + Date::ORDINAL_STEP; ;
			 d += Date::ORDINAL_STEP )
		{
			CategoryChapters::iterator iter( m_topics.find( d ) );
			if( iter == m_topics.end() )
				break;
			Chapter *chpt( iter->second );
			m_topics.erase( d );
			chpt->set_date( Date( d - Date::ORDINAL_STEP ) );
			m_topics[ chpt->get_date().m_date ] = chpt;
		}

		// SHIFT ENTRIES
		if( m_entries.size() > 0 )
		{
			for( Date::date_t d = d_chapter.m_date + 1; d <= m_entries.begin()->first; )
			{
				Entryiter iter( m_entries.find( d ) );
				if( iter == m_entries.end() )
				{
					d += Date::ORDINAL_STEP;
					Date::reset_order_1( d );
					continue;
				}

				Entry *entry( iter->second );
				int order_diff( entry->get_date().get_ordinal_order() -
								d_chapter.get_ordinal_order() );

				if( order_diff > 0 || ( order_diff == 0 && flag_last_chapter ) )
				{
					m_entries.erase( d );
					Date d_new( d - Date::ORDINAL_STEP );
					if( order_diff == 1 || ( order_diff == 0 && flag_last_chapter ) )
						d_new.m_date += last_order_of_prev_chapter;
					entry->set_date( d_new );
					m_entries[ d_new.m_date ] = entry;
				}
				d++;
			}
		}
	}
	else
		m_ptr2chapter_ctg_cur->dismiss_chapter( chapter );
}

bool
Diary::rename_chapter_ctg( CategoryChapters *category, const Glib::ustring &name )
{
	if( m_chapter_categories.count( name ) > 0 )
		return false;

	m_chapter_categories.erase( category->get_name() );
	category->set_name( name );
	m_chapter_categories.insert(
			PoolCategoriesChapters::value_type( name, category ) );

	return true;
}

Chapter*
Diary::add_topic( const Glib::ustring &name )
{
	Chapter *c( m_topics.create_chapter( name, m_topics.get_free_order_ordinal() ) );
	c->set_expanded( true );
	return c;
}

Date
Diary::get_free_chapter_order_temporal( void )
{
	time_t t = time( NULL );
	struct tm *ti = localtime( &t );
	Date date( ti->tm_year + 1900, ti->tm_mon + 1, ti->tm_mday );

	while( m_ptr2chapter_ctg_cur->get_chapter( date.m_date ) )
	{
		date.forward_day();
	}

	return date;
}

Date
Diary::get_free_chapter_order_ordinal( void )
{
	return m_ptr2chapter_ctg_cur->get_free_order_ordinal();
}

bool
Diary::make_free_entry_order( Date &date )
{
	date.reset_order_1();
	while( m_entries.find( date.m_date ) != m_entries.end() )
		++date.m_date;

	return true;	// reserved for bounds checking
}

// THEMES ==========================================================================================
Theme*
Diary::create_theme( const Glib::ustring& name )
{
    // system theme added back to m_themes:
//    if( name == STRING::SYSTEM_THEME )
//        return ThemeSystem::get();

	PoolThemes::iterator iter = m_themes.find( name );
	if( iter != m_themes.end() )
		return( iter->second );
	Theme *theme = new Theme( this, name );
	m_themes.insert( PoolThemes::value_type( name, theme ) );
	return theme;
}

Theme*
Diary::duplicate_theme( const Theme *theme )
{
	Glib::ustring name = create_unique_name_for_map( m_themes, _( "New theme" ) );
	Theme *new_theme = new Theme( theme );	// duplication
	new_theme->m_name = name;
	m_themes.insert( PoolThemes::value_type( name, new_theme ) );

	return new_theme;
}

void
Diary::dismiss_theme( Theme *theme )
{
	// change default theme if necessary
	if( m_default_theme == theme )
	{
		m_default_theme = ThemeSystem::get();
	}
	// fix entries
	for( Entryiter iter = m_entries.begin(); iter != m_entries.end(); ++iter )
		if( iter->second->m_ptr2theme == theme )
			iter->second->m_ptr2theme = NULL;
	// remove from the list and delete
	m_themes.erase( theme->get_name() );
	delete theme;
}

void
Diary::show( void )
{
	if( shower != NULL )
		shower->show( *this );
}

// IMPORTING =======================================================================================
bool
Diary::import_tag( Tag *tag )
{
	return( create_tag( tag->get_name() ) != NULL );
}

bool
Diary::import_entryset( const PoolEntries &set,
						bool flag_import_tags,
						const Glib::ustring &tag_all_str )
{
	Entry *entry_new, *entry_ext;
	Tag *tag_all = NULL;

	if( ! tag_all_str.empty() )
		tag_all = create_tag( tag_all_str );

	for( Entryciter iter = set.begin();
		 iter != set.end();
		 ++iter )
	{
		entry_ext = iter->second;
		entry_new = new Entry( this,
							   entry_ext->m_date.m_date,
							   entry_ext->m_text,
							   entry_ext->m_option_favored );

		// fix order:
		entry_new->m_date.reset_order_1();
		while( m_entries.find( entry_new->m_date.m_date ) != m_entries.end() )
			entry_new->m_date.m_date++;

        // try to preserve the theme:
        if( entry_ext->get_theme_is_set() )
        {
            PoolThemes::iterator iter( m_themes.find( entry_ext->get_theme()->get_name() ) );
            if( iter != m_themes.end() )
                entry_new->set_theme( iter->second );
        }

		// copy dates:
		entry_new->m_date_created = entry_ext->m_date_created;
		entry_new->m_date_changed = entry_ext->m_date_changed;

		// insert it into the diary:
		m_entries[ entry_new->m_date.m_date ] = entry_new;

		if( flag_import_tags )
		{
			Tagset &tags = entry_ext->get_tags();
			Tag *tag;

			for( Tagset::const_iterator iter = tags.begin(); iter != tags.end(); ++iter )
			{
				tag = ( m_tags.find( ( *iter )->get_name() ) )->second;
				entry_new->add_tag( tag );
			}
		}
		if( tag_all )
			entry_new->add_tag( tag_all );
	}

	return true;
}

bool
Diary::import_themes( const PoolThemes &themes )
{
    for( PoolThemes::const_iterator iter = themes.begin();
         iter != themes.end();
         ++iter )
    {
        Theme *theme( iter->second );
        if( theme->is_system() )
            continue;

        Glib::ustring name( create_unique_name_for_map( m_themes, theme->m_name ) );
        Theme *new_theme( new Theme( theme ) ); // theme is duplicated
        new_theme->m_name = name;
        m_themes.insert( PoolThemes::value_type( name, new_theme ) );
    }

    return true;
}
