/***********************************************************************************

	Copyright (C) 2007-2011 Ahmet Öztürk (aoz_2@yahoo.com)

	This file is part of Lifeograph.

	Lifeograph is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Lifeograph is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

***********************************************************************************/


#ifndef LIFEOGRAPH_DIARY_HEADER
#define LIFEOGRAPH_DIARY_HEADER


#include "helpers.hpp"	// i18n headers
#include "diarydata.hpp"
#include "entry.hpp"


namespace LIFEO
{
const char						DB_FILE_HEADER[] = "LIFEOGRAPHDB";
const int						DB_FILE_VERSION_INT( 110 );
const int						DB_FILE_VERSION_INT_MIN( 56 );
const std::string::size_type	PASSPHRASE_MIN_SIZE( 4 );
const char						LOCK_SUFFIX[] = ".~LOCK~";


enum Result
{
	OK,
	ABORTED,
	SUCCESS,
	FAILURE,
	COULD_NOT_START,
	COULD_NOT_FINISH,
	WRONG_PASSWORD,
	APPEARENTLY_ENCRYTED_FILE,
	APPEARENTLY_PLAIN_FILE,
	INCOMPATIBLE_FILE,
	CORRUPT_FILE,
	// EMPTY_DATABASE, // not used anymore
	// VARIABLES USED BY set_path():
	FILE_NOT_FOUND,
	FILE_NOT_READABLE,
	FILE_LOCKED
};


typedef PoolEntries::iterator				Entryiter;
typedef PoolEntries::reverse_iterator		EntryIterReverse;
typedef PoolEntries::const_iterator			Entryciter;

typedef std::vector< Entry* >				Entryvector;
typedef std::vector< Entry* >::iterator		Entryiterv;


class Diary : public DiaryBase, public DiaryElement
{
	public:
        typedef unsigned int FilteringStatus;
        static const FilteringStatus FS_CLEAR               = 0;
        static const FilteringStatus FS_FILTER_TEXT         = 2;
        static const FilteringStatus FS_FILTER_FAVORITES    = 4;
        static const FilteringStatus FS_FILTER_TAG          = 8;
        static const FilteringStatus FS_FILTER_DATE         = 16;
        static const FilteringStatus FS_FILTER_TRASH        = 32;
        static const FilteringStatus FS_NEW                 = 1024;

		static ElementShower< Diary > *shower;

								Diary( void );
								~Diary( void );

		static Diary			*d;

		bool					is_old( void )
		{ return( m_read_version < DB_FILE_VERSION_INT ); }

		void					init_new( const std::string& );

		// DISK I/O
		LIFEO::Result			read_body( void );
		LIFEO::Result			read_header( void );

		LIFEO::Result			write( void );
		LIFEO::Result			write( const std::string& );
		LIFEO::Result			write_copy( const std::string&, const std::string&, bool );
		LIFEO::Result			write_txt( const std::string&, bool );

		virtual void			clear( void );

		int						get_size() const
		{ return m_entries.size(); }
		Type					get_type( void ) const
		{ return ET_DIARY; }
		const Glib::RefPtr< Gdk::Pixbuf >&
								get_icon( void ) const;
        const Glib::RefPtr< Gdk::Pixbuf >&
                                get_icon32( void ) const;

		LIFEO::Result			set_path( const std::string&, bool, bool = false );
		const std::string&		get_path( void ) const;
		bool					is_path_set( void ) const;

		bool					set_passphrase( const std::string& );
		void					clear_passphrase( void );
		std::string				get_passphrase( void ) const;
		bool					compare_passphrase( const std::string& ) const;
		bool					is_passphrase_set( void ) const;

        bool                    is_read_only( void ) const
        { return m_flag_read_only; }

		DiaryElement*			get_element( DEID );
		DiaryElement*			get_startup_elem( void );
		DiaryElement*			get_most_current_elem( void );
		DiaryElement*			get_prev_session_elem( void );

		Entry*					get_entry( const Date::date_t );
		Entryvector*			get_entries( Date::date_t );
		PoolEntries&			get_entries( void )
		{ return m_entries; }
		bool					get_day_has_multiple_entries( const Date& );
		Entry*					get_entry_today( void );
		Entry*					get_entry_next_in_day( const Date& );
		Entry*					get_entry_first( void );
		void					set_entry_date( Entry*, const Date& );
		//int						get_entrydayorder( const Entry * const ) const;

		bool					is_first( const Entry * const ) const;
		bool					is_last( const Entry * const ) const;
		SortingCriteria			get_sorting_criteria( void ) const
		{ return m_option_sorting_criteria; }
		void					set_sorting_criteria( SortingCriteria sc )
		{ m_option_sorting_criteria = sc; }

		void					set_filter_text( const Glib::ustring& );
		Glib::ustring			get_filter_text( void ) const
		{ return m_filter_text; }
		void					set_filter_tag( const Tag* );
		const Tag*				get_filter_tag( void ) const
		{ return m_filter_tag; }
        void                    toggle_filter_favorites( void );
        void                    toggle_filter_trash( void );
        Date::date_t            get_filter_date_begin( void )
        { return m_filter_date_begin; }
        Date::date_t            get_filter_date_end( void )
        { return m_filter_date_end; }
        void                    set_filter_date_begin( Date::date_t d );
        void                    set_filter_date_end( Date::date_t d );
        void                    activate_individual_filter( void );
        void                    clear_filter_date( void );

        FilteringStatus         get_filtering_status( void ) const { return m_filtering_status; }
		void					set_filtering_status_applied( void )
		{	if( m_filtering_status & FS_NEW ) m_filtering_status -= FS_NEW; }
        void                    set_filtering_status_new( void ) { m_filtering_status |= FS_NEW; }

		int						replace_text( const Glib::ustring& );

        Entry*					create_entry( Date::date_t, const Glib::ustring& = "",
                                              bool = false );
		// adds a new entry to today even if there is already one or more:
		Entry*					add_today( void );
		virtual bool			dismiss_entry( Entry* );
		int						list_entries( void );

		PoolTags*				get_tags( void )
		{ return &m_tags; }
		CategoryTags*			create_tag_ctg( void );
		CategoryTags*			create_tag_ctg( const Glib::ustring& );
		void					dismiss_tag_ctg( CategoryTags* );
		Tag*					create_tag( const Glib::ustring&, CategoryTags* = NULL );
		void					dismiss_tag( Tag* );

		const PoolCategoriesChapters*
								get_chapter_ctgs( void ) const
		{ return &m_chapter_categories; }
		CategoryChapters*		get_current_chapter_ctg( void )
		{ return m_ptr2chapter_ctg_cur; }
		CategoryChapters*		create_chapter_ctg( void );
		CategoryChapters*		create_chapter_ctg( const Glib::ustring& );
		void					dismiss_chapter_ctg( CategoryChapters* );
		void					dismiss_chapter( Chapter* );
		bool					rename_chapter_ctg( CategoryChapters*, const Glib::ustring& );
		const CategoryChapters*	get_topics( void )
		{ return &m_topics; }
		Chapter*				add_topic( const Glib::ustring& );

		Date					get_free_chapter_order_temporal( void );
		Date					get_free_chapter_order_ordinal( void );

		bool					make_free_entry_order( Date& );

		bool					is_encrypted( void ) const
		{ return( ! m_passphrase.empty() ); }

		Theme*					create_theme( const Glib::ustring& );
		Theme*					duplicate_theme( const Theme *theme );
		bool					has_theme( void ) const
		{ return( m_default_theme != NULL ); }
		const PoolThemes&		get_themes( void ) const
		{ return m_themes; }
		PoolThemes&				get_themes( void )
		{ return m_themes; }
		void					set_default_theme( Theme *theme )
		{ m_default_theme = theme; }
		Theme*					get_default_theme( void ) const
		{ return m_default_theme; }
		void					dismiss_theme( Theme *theme );

		void					show( void );

		bool					import_tag( Tag* );
		bool					import_entryset( const PoolEntries&, bool, const Glib::ustring& );
		bool					import_themes( const PoolThemes& );

	protected:
		std::string				m_path;
		std::string				m_passphrase;
		// content
		PoolEntries				m_entries;
		PoolTags				m_tags;
		PoolCategoriesTags		m_tag_categories;
		PoolCategoriesChapters	m_chapter_categories;
		CategoryChapters		*m_ptr2chapter_ctg_cur;
		CategoryChapters		m_topics;	// ordinal chapters
		PoolThemes				m_themes;
		Theme					*m_default_theme;

		DEID					m_startup_elem;
		DEID					m_last_elem;
		// options & flags
		SortingCriteria			m_option_sorting_criteria;
		int						m_read_version;
		bool                    m_flag_only_save_filtered;
		bool                    m_flag_ignore_locks;
		bool                    m_flag_read_only;
		// filtering
		Glib::ustring			m_filter_text;
		const Tag				*m_filter_tag;
        Date::date_t            m_filter_date_begin;
        Date::date_t            m_filter_date_end;
		FilteringStatus			m_filtering_status;

		LIFEO::Result			parse_db_body_text( std::stringstream& );
		LIFEO::Result			parse_db_body_text_56( std::stringstream& );
		bool					create_db_body_text( std::stringstream& );
		bool					create_db_header_text( std::stringstream&, bool );
		LIFEO::Result			read_plain( void );
		LIFEO::Result			read_encrypted( void );
		LIFEO::Result			write_plain( const std::string&, bool = false );
		LIFEO::Result			write_encrypted( const std::string& );

		bool                    remove_lock();

	private:
		std::ios::pos_type		m_body_position;

	friend class Lifeograph;
	friend class EntryView;
	friend class DiaryView;
	friend class TagView;
	friend class CategoryTagsView;
	friend class ThemeView;
	friend class PanelDiary;
	friend class PanelExtra;
	friend class WidgetPanelExp;
	friend class DialogImport;
};

} // end of namespace LIFEO

#endif

